/* ezcdb_cross.c
** ezcdb cross module
** wcm, 2010.11.02 - 2010.11.24
** ===
*/

/* libc: */
#include <stdlib.h>
/* prototype for rename() in stdio.h: */
extern int     rename(const char *from, const char *to);

/* unix: */
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "nuscan.h"
#include "sysstr.h"
#include "uchar.h"
#include "upak.h"

/* lasagna cdb reader: */
#include "cdb.h"
/* lasagna cdb creator: */
#include "cdbmk.h"

/* ezcdb: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_cross
#endif                          /* #ifndef EZCDB_MODULE */


/*
** objects in local scope - options:
*/


/*
** objects in local scope - other:
*/
static const char *tmpext = ".{new}";
/* M = A intersect B */
struct cdbmk   M;
struct cdb     cdb_A, cdb_B;

/*
** functions in local scope:
*/
static int     do_cross(void);


/* do_cross()
**   into initialized cdbmk object M, intersection of A and B
**   ie, output records from A with matching keys in B
**   abort on failure
*/
static
int do_cross()
{
   struct cdb    *A, *B;
   uint32_t       kpos, klen;
   dynbuf         K = dynbuf_INIT();
   int            r;

   A = &cdb_A;
   B = &cdb_B;

   /* scan A records: */
   for (r = cdb_seqinit(A); r != 0; r = cdb_seqnext(A)) {
      if (r == -1) {
         fatal_syserr("failure during cdb_seq()");
      }

      /* only include records from A with key found in B: */
      klen = cdb_klen(A);
      kpos = cdb_kpos(A);

      /* XXX, optimize for mmap()'d A */
      dynbuf_CLEAR(&K);
      if (cdb_dynread(A, &K, klen, kpos) == -1) {
         fatal_syserr("failure during cdb_dynread()");
      }

      /* lookup A key in B: */
      switch (cdb_find(B, dynbuf_BUF(&K), klen)) {
      case -1:
         fatal_syserr("failure during cdb_find()");
         break;
      case 0:
         /* no match, skip this record in A: */
         continue;
      }

      /* output current record in A: */
      if (cdbmk_addrec(&M, A) == -1) {
         fatal_syserr("failure during cdbmk_addrec()");
      }
   }

   /* finish make: */
   if (cdbmk_finish(&M) == -1) {
      fatal_syserr("failure during cdbmk_finish()");
   }

   /* all done: */
   return 0;
}


int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   uint32_t       u;
   const char    *z;
   int            fd = -1, perms = -1;
   char          *outfile = NULL, *tmpfile = NULL;
   char          *fn_tmp = NULL;        /* non-NULL if allocated */
   char          *cdbA_file, *cdbB_file;
   int            e;

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-p perms] [-t tmpfile] result.cdb A.cdb B.cdb";
   nextopt_init(&nopt, argc, argv, ":hVap:t:");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-p perms] [-t tmpfile] result.cdb A.cdb B.cdb";
   nextopt_init(&nopt, argc, argv, ":hap:t:");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'p':
         z = nuscan_uint32o(&u, nopt.opt_arg);
         if (*z != '\0') {
            fatal_usage("non-numeric/octal argument found for option -", optc,
                        ": ", nopt.opt_arg);
         }
         perms = (int) u;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      /* result_file argument: */
      outfile = argv[0];
      ++argv;
      --argc;
   } else {
      fatal_usage("missing \"result.cdb\" file argument");
   }

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      /* cdbA_file argument: */
      cdbA_file = argv[0];
      ++argv;
      --argc;
   } else {
      fatal_usage("missing \"A.cdb\" file argument");
   }

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      /* cdbB_file argument: */
      cdbB_file = argv[0];
      ++argv;
      --argc;
   } else {
      fatal_usage("missing \"B.cdb\" file argument");
   }

   if (tmpfile == NULL) {
      fn_tmp = (char *) malloc(cstr_len(outfile) + cstr_len(tmpext) + 1);
      if (fn_tmp == NULL) {
         fatal_syserr("failure malloc() for naming tmpfile");
      }
      cstr_vcopy(fn_tmp, outfile, tmpext);
      tmpfile = fn_tmp;
   }

   if (perms >= 0)
      umask(0);
   else
      perms = 0666;

   unlink(tmpfile);
   fd = open(tmpfile, O_WRONLY | O_CREAT | O_EXCL, (mode_t) perms);
   if (fd == -1) {
      fatal_syserr("failure on open() for ", tmpfile);
   }

   if (cdbmk_init(&M, fd) == -1) {
      fatal_syserr("failure cdbmk_init()");
   }

   if (cdb_open(&cdb_A, cdbA_file) == -1) {
      fatal_syserr("failure cdb_open() on ", cdbA_file);
   }

   if (cdb_open(&cdb_B, cdbB_file) == -1) {
      fatal_syserr("failure cdb_open() on ", cdbB_file);
   }

   e = do_cross();

   if (e == 0) {
      if (fsync(fd) == -1) {
         fatal_syserr("failure fsync()");
      }

      if (close(fd) == -1) {
         fatal_syserr("failure close()");
      }

      if (rename(tmpfile, outfile) == -1) {
         fatal_syserr("failure rename()");
      }
   }

   /* gratuitous cleanup: */
   cdb_close(&cdb_A);
   cdb_close(&cdb_B);
   cdbmk_clear(&M);
   if (fn_tmp != NULL)
      free(fn_tmp);

   return e;
}


/* eof (ezcdb_cross.c) */
