/* ezcdb_dump.c
** ezcdb dump: write out contents of cdb file
** wcm, 2010.05.30 - 2010.12.01
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "die.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"

/* lasagna cdb reader: */
#include "cdb.h"

/* ezcdb: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_dump
#endif                          /* #ifndef EZCDB_MODULE */


/* objects in local scope: */
static uchar_t *map = NULL;
static uint32_t mapsize = 0;
static char    nfmt[NFMT_SIZE];
#define  DUMPBUF  256
static uchar_t dumpbuf[DUMPBUF];

/* -g: "getln" output mode */
static int     opt_g = 0;
/* -x: "xchange" hdb output mode */
static int     opt_x = 0;
/* -d arg: delimiter for -g mode */
static char   *arg_d = "\t";


/* functions in local scope: */
static void    do_put_map(struct cdb *C, uint32_t offset, size_t len);
static void    do_put_buf(struct cdb *C, uint32_t offset, size_t len);
static int     do_dump(struct cdb *C);


/* do_put_map()
**   direct copy from non-NULL map to ioq1
**   abort on failure
*/
static
void do_put_map(struct cdb *NOTUSED, uint32_t offset, size_t len)
{
   /* silence compiler warning unused parameter (nv): */
   (void) NOTUSED;

   if ((offset > mapsize) || ((mapsize - offset) < len)) {
      errno = EPROTO;
      fatal_syserr("map offset request exceeds map boundary");
   }
   ioq_put(ioq1, map + offset, len);

   return;
}

/* do_put_buf()
**   copy from cdb to ioq1 through intermediate dumpbuf
**   abort on failure
*/
static
void do_put_buf(struct cdb *C, uint32_t offset, size_t len)
{
   size_t         n = sizeof dumpbuf;

   while (len) {
      if (len < n)
         n = len;
      if (cdb_read(C, dumpbuf, n, offset) == -1) {
         fatal_syserr("failure during cdb_read()");
      }
      ioq_put(ioq1, dumpbuf, n);
      offset += n;
      len -= n;
   }

   return;
}


/* do_dump()
**   iterate and output each record in cdb 
*/
static
int do_dump(struct cdb *C)
{
   uint32_t       klen, dlen;
   uint32_t       rpos, offset;
   void           (*do_put) (struct cdb *, uint32_t, size_t);
   int            r;

   /* setup for either direct map or intermediate dumpbuf operation: */
   map = cdb_map(C);
   if (map) {
      mapsize = cdb_mapsize(C);
      do_put = &do_put_map;
   } else {
      do_put = &do_put_buf;
   }

   /* sequential iteration: */
   for (r = cdb_seqinit(C); r != 0; r = cdb_seqnext(C)) {
      if (r == -1) {
         fatal_syserr("failure during cdb_seq()");
      }

      klen = cdb_klen(C);
      dlen = cdb_dlen(C);
      rpos = cdb_rpos(C);

      if (!opt_g) {
         if (!opt_x)
            ioq_puts(ioq1, "+");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), opt_x ? ":" : ",");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, dlen), opt_x ? "\t" : ":");
      }

      offset = rpos + 8;
      do_put(C, offset, klen);
      if (!opt_g) {
         ioq_puts(ioq1, opt_x ? ":" : "->");
      } else {
         ioq_put(ioq1, (uchar_t *) arg_d, 1);
      }

      offset += klen;
      do_put(C, offset, dlen);
      ioq_puts(ioq1, "\n");
   }

   if (!opt_g && !opt_x)
      ioq_puts(ioq1, "\n");

   ioq_flush(ioq1);
   return 0;
}


int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct cdb     C;
   int            is_stdin = 0;
   int            e;

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-d del | -g | -x] [cdb]";
   nextopt_init(&nopt, argc, argv, ":hVd:gx");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-d del | -g | -x] [cdb]";
   nextopt_init(&nopt, argc, argv, ":hcd:gx");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'g':
         ++opt_g;
         break;
      case 'd':
         ++opt_g;
         arg_d = nopt.opt_arg;
         break;
      case 'x':
         ++opt_x;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_g && opt_x) {
      fatal_usage("options -g and -x are mutually exclusive");
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* cdb file argument: */
      if (cdb_open(&C, argv[0]) == -1) {
         fatal_syserr("failure initializing cdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (cdb_init(&C, 0) == -1) {
         fatal_syserr("failure initializing cdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_dump() just aborts on failure */
   e = do_dump(&C);

   /* close/cleanup: */
   if (is_stdin) {
      cdb_clear(&C);
   } else {
      cdb_close(&C);
   }

   return e;
}


/* eof (ezcdb_dump.c) */
