/* ezcdb_get.c
** ezcdb get: find key in cdb
** wcm, 2010.05.30 - 2010.11.22
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "cstr.h"
#include "dynbuf.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nuscan.h"
#include "sysstr.h"
#include "uchar.h"

/* lasagna cdb reader: */
#include "cdb.h"

/* ezcdb: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_get
#endif                          /* #ifndef EZCDB_MODULE */


/* objects in local scope: */
/* -a: "all" */
static int     opt_a = 0;
/* -n: "omit trailing newline" */
static int     opt_n = 0;
/* -j arg: "jump" */
static int     opt_j = 0;
static uint32_t arg_j = 0;


/* functions in local scope: */
static int     do_get(struct cdb *C, const char *key);


/* do_get()
**   lookup key in cdb
**   return:
**     0: found
**     1: not found
**
**   abort on failure
*/
static
int do_get(struct cdb *C, const char *key)
{
   uint32_t       klen;
   dynbuf         D = dynbuf_INIT();
   int            found;
   uint32_t       jump = opt_j ? arg_j : 0;
   uint32_t       n;

   klen = cstr_len(key);

   for (found = cdb_find(C, (uchar_t *) key, klen), n = 0;
        found > 0; found = cdb_findnext(C), ++n) {
      /* show conditions: */
      if (opt_a || (opt_j && (n == jump)) || (!opt_a && !opt_j && (n == 0))) {
         dynbuf_CLEAR(&D);
         if (cdb_dynget(C, &D) == -1) {
            fatal_syserr("failure cdb_dynget()");
         }
         ioq_put(ioq1, dynbuf_BUF(&D), dynbuf_LEN(&D));
         if (!opt_n) {
            ioq_puts(ioq1, "\n");
         }
      }

      /* stop conditions: */
      if ((opt_j && (n == jump)) || (!opt_a && !opt_j && (n == 0))) {
         break;
      }
   }
   ioq_flush(ioq1);

   if (found == -1) {
      fatal_syserr("failure during cdb_find()");
   }

   if (found || (opt_a && n)) {
      return 0;
   }

   /* else key not found: */
   return 1;
}


int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct cdb     C;
   char          *key;
   const char    *z;
   int            is_stdin = 0;
   int            e;

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-a | -j num] [-n] key [cdb]";
   nextopt_init(&nopt, argc, argv, ":hVaj:n");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-a | -j num] [-n] key [cdb]";
   nextopt_init(&nopt, argc, argv, ":haj:n");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'a':
         ++opt_a;
         break;
      case 'j':
         ++opt_j;
         z = nuscan_uint32(&arg_j, nopt.opt_arg);
         if (*z != '\0') {
            fatal_usage("non-numeric argument found for option -", optc, ": ",
                        nopt.opt_arg);
         }
         break;
      case 'n':
         ++opt_n;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_a && opt_j) {
      fatal_usage("options -a and -j are mutually exclusive");
   }

   if ((argc < 1) || ((key = argv[0]) == NULL)) {
      fatal_usage("missing search key argument");
   }
   --argc;
   ++argv;

   if ((argc > 0) && (argv[0] != NULL)) {
      /* cdb file argument: */
      if (cdb_open(&C, argv[0]) == -1) {
         fatal_syserr("failure initializing cdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (cdb_init(&C, 0) == -1) {
         fatal_syserr("failure initializing cdb on stdin");
      }
      ++is_stdin;
   }


   /* do_get() returns 0 if found, 1 if not found; aborts on error */
   e = do_get(&C, key);

   /* close/cleanup: */
   if (is_stdin) {
      cdb_clear(&C);
   } else {
      cdb_close(&C);
   }

   return e;
}


/* eof (ezcdb_get.c) */
