/* ezcdb_grep.c
** ezcdb grep: regex match scan through cdb file
** wcm, 2010.11.01 - 2010.11.22
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* regex: */
#include <regex.h>

/* lasagna: */
#include "cstr.h"
#include "die.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"

/* lasagna cdb reader: */
#include "cdb.h"

/* ezcdb: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_grep
#endif                          /* #ifndef EZCDB_MODULE */

/*
** objects in local scope - options:
*/

/* -i: case insensitive match */
static int     opt_i = 0;
/* -k: match on keys */
static int     opt_k = 0;
/* -!: invert match */
static int     opt_NOT = 0;

/* -g: "getline" output mode */
static int     opt_g = 0;
/* -x: "xchange" hdb output mode */
static int     opt_x = 0;
/* -d arg: delimiter for -g mode */
static char   *arg_d = "\t";

/*
** objects in local scope - other :
*/

/* compiled regex: */
static regex_t rx;

/* stringified numerics: */
static char    nfmt[NFMT_SIZE];

/* functions in local scope: */
static int     do_grep(struct cdb *C);


/* do_grep()
**   iterate and grep each record
**   abort on fail
**
**   return:
**     0 : success, 1 or more matches found
**     1 : no match found
*/
static
int do_grep(struct cdb *C)
{
   uint32_t       klen, dlen;
   dynbuf         K = dynbuf_INIT();
   dynbuf         D = dynbuf_INIT();
   char          *key, *data;
   uint32_t       match_found = 0;
   int            r, e;

   /* sequential iteration: */
   for (r = cdb_seqinit(C); r != 0; r = cdb_seqnext(C)) {
      if (r == -1) {
         fatal_syserr("failure during cdb_seq()");
      }
      klen = cdb_klen(C);
      dlen = cdb_dlen(C);

      dynbuf_CLEAR(&K);
      if (cdb_dynread(C, &K, klen, cdb_kpos(C))) {
         fatal_syserr("failure during cdb_dynread()");
      }
      dynbuf_putnul(&K);
      key = (char *) dynbuf_BUF(&K);

      dynbuf_CLEAR(&D);
      if (cdb_dynread(C, &D, dlen, cdb_dpos(C))) {
         fatal_syserr("failure during cdb_dynread()");
      }
      dynbuf_putnul(&D);
      data = (char *) dynbuf_BUF(&D);

      e = regexec(&rx, (opt_k ? key : data), 0, NULL, 0);

      /* conditions skipping output for this record: */
      if (                      /* no match, not inverted: */
            ((e == REG_NOMATCH) && !opt_NOT) ||
            /* match, but inverted: */
            ((e == 0) && opt_NOT)) {
         continue;
      } else if (e) {
         /* XXX, error regexec() */
         warn("failure regexec() on ", opt_k ? "key: " : "data: ",
              opt_k ? key : data);
         continue;
      }

      /* else match: */
      ++match_found;

      /* start output: */
      if (!opt_g) {
         if (!opt_x)
            ioq_puts(ioq1, "+");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), opt_x ? ":" : ",");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, dlen), opt_x ? "\t" : ":");
      }

      /* output key: */
      ioq_put(ioq1, (uchar_t *) key, klen);

      /* output delimiter: */
      if (!opt_g) {
         ioq_puts(ioq1, opt_x ? ":" : "->");
      } else {
         ioq_put(ioq1, (uchar_t *) arg_d, 1);
      }

      /* output data: */
      ioq_put(ioq1, (uchar_t *) data, dlen);

      /* record terminal: */
      ioq_puts(ioq1, "\n");
   }

   /* empty line terminal for cdb output: */
   if (!opt_g && !opt_x)
      ioq_puts(ioq1, "\n");

   ioq_flush(ioq1);

   dynbuf_freebuf(&K);
   dynbuf_freebuf(&D);

   return (match_found ? 0 : 1);
}


int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   const char    *regex = NULL;
   int            rxflags = 0;
   struct cdb     C;
   int            is_stdin = 0;
   int            e;

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-d del | -g | -x] [-i] [-k] [-!] regex [cdb]";
   nextopt_init(&nopt, argc, argv, ":hVd:gxik!");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-d del | -g | -x] [-i] [-k] [-!] regex [cdb]";
   nextopt_init(&nopt, argc, argv, ":hd:gxik!");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case '!':
         ++opt_NOT;
         break;
      case 'd':
         ++opt_g;
         arg_d = nopt.opt_arg;
         break;
      case 'g':
         ++opt_g;
         break;
      case 'i':
         ++opt_i;
         break;
      case 'k':
         ++opt_k;
         break;
      case 'x':
         ++opt_x;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      regex = argv[0];
      --argc;
      ++argv;
   } else {
      fatal_usage("missing or empty regex argument");
   }

   if (opt_g && opt_x) {
      fatal_usage("options -g and -x are mutually exclusive");
   }

   rxflags = REG_EXTENDED | REG_NOSUB;
   if (opt_i)
      rxflags |= REG_ICASE;

   /* compile regex: */
   e = regcomp(&rx, regex, rxflags);
   if (e != 0) {
      switch (e) {
      case REG_ESPACE:
         fatal(111, "failure compiling regex: out of memory");
         break;
      default:
         fatal_usage("bad regex pattern");
      }
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* cdb file argument: */
      if (cdb_open(&C, argv[0]) == -1) {
         fatal_syserr("failure initializing cdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (cdb_init(&C, 0) == -1) {
         fatal_syserr("failure initializing cdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_grep() just aborts on failure */
   e = do_grep(&C);

   /* cleanup: */
   regfree(&rx);
   if (is_stdin) {
      cdb_clear(&C);
   } else {
      cdb_close(&C);
   }

   return e;
}


/* eof (ezcdb_grep.c) */
