/* ezcdb_keys.c
** ezcdb keys: write out keys in cdb file
** wcm, 2010.05.30 - 2010.11.18
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "die.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"

/* lasagna cdb reader: */
#include "cdb.h"

/* ezcdb: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_keys
#endif                          /* #ifndef EZCDB_MODULE */


/* objects in local scope: */
static char    nfmt[NFMT_SIZE];

/* -g: "getline" output mode */
static int     opt_g = 0;
/* -x: "xchange" hdb output mode */
static int     opt_x = 0;
/* -X: display hash (in hex) */
static int     opt_X = 0;

/* functions in local scope: */
static int     do_keys(struct cdb *C);


/* do_keys()
**   iterate and output each key in cdb 
**   abort on failure
*/
static
int do_keys(struct cdb *C)
{
   uint32_t       klen, dlen, rpos;
   dynbuf         K = dynbuf_INIT();
   uchar_t       *key;
   int            r;

   /* sequential iteration: */
   for (r = cdb_seqinit(C); r != 0; r = cdb_seqnext(C)) {
      if (r == -1) {
         fatal_syserr("failure during cdb_seq()");
      }

      klen = cdb_klen(C);
      dlen = cdb_dlen(C);
      rpos = cdb_rpos(C);

      dynbuf_CLEAR(&K);
      if (cdb_dynread(C, &K, klen, rpos + 8)) {
         fatal_syserr("failure during cdb_dynread()");
      }
      key = (uchar_t *) dynbuf_BUF(&K);

      /* output klen: */
      if (!opt_g) {
         if (!opt_x)
            ioq_puts(ioq1, "+");
         ioq_vputs(ioq1, nfmt_uint32(nfmt, klen), ":");
      }

      /* output key: */
      ioq_put(ioq1, key, klen);

      /* output hash: */
      if (opt_X) {
         uint32_t       hash = cdb_hash(dynbuf_BUF(&K), klen);
         ioq_vputs(ioq1, "\t[0x", nfmt_uint32x_pad0(nfmt, hash, 8), "]");
      }
      ioq_puts(ioq1, "\n");
   }

   /* empty line terminal for cdb output: */
   if (!opt_g && !opt_x)
      ioq_puts(ioq1, "\n");

   ioq_flush(ioq1);

   dynbuf_freebuf(&K);
   return 0;
}


int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct cdb     C;
   int            is_stdin = 0;
   int            e;

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-g | -x] [-X] [cdb]";
   nextopt_init(&nopt, argc, argv, ":hVgxX");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-g | -x] [-X] [cdb]";
   nextopt_init(&nopt, argc, argv, ":hgxX");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'g':
         ++opt_g;
         break;
      case 'x':
         ++opt_x;
         break;
      case 'X':
         ++opt_X;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_g && opt_x) {
      fatal_usage("options -g and -x are mutually exclusive");
   }

   if ((argc > 0) && (argv[0] != NULL)) {
      /* cdb file argument: */
      if (cdb_open(&C, argv[0]) == -1) {
         fatal_syserr("failure initializing cdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (cdb_init(&C, 0) == -1) {
         fatal_syserr("failure initializing cdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_keys() just aborts on failure */
   e = do_keys(&C);

   /* close/cleanup: */
   if (is_stdin) {
      cdb_clear(&C);
   } else {
      cdb_close(&C);
   }

   return e;
}


/* eof (ezcdb_keys.c) */
