/* ezcdb_make.c
** ezcdb make: cdb generator utility
** wcm, 2010.05.30 - 2010.11.19
** ===
*/

/* libc: */
#include <stdlib.h>
/* prototype for rename() in stdio.h: */
extern int     rename(const char *from, const char *to);

/* unix: */
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "dynbuf.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "nuscan.h"
#include "sysstr.h"
#include "uchar.h"

/* lasagna cdb generator/writer: */
#include "cdbmk.h"

/* ezcdb applications header: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_make
#endif                          /* #ifndef EZCDB_MODULE */


/* objects in local scope: */
static const char *tmpext = ".{new}";
static int     eof = 0, eof_ok = 0;
static size_t  input_line = 0;
static char    nfmt[NFMT_SIZE];

/* data input formats supported:
**
**   cdb input format:       [+klen,dlen:key->data\n] (default)
**
**   cdb "getln" format:     [key\tdata\n] 
**
**   hdb "xchange" format:   [klen:dlen\tkey:data\n]
**
**   notes:
**     "getln" format supports commented lines
*/

/* -g: "getln" input format: [key\tvalue] */
static int     opt_g = 0;
/* -x: "xchange" input format: [klen:dlen\tkey:data\n] */
static int     opt_x = 0;
/* -d arg: key/value delimiter for -g format (implies -g)*/
static char   *arg_d = "\t";


/* additional die() macros: */
#define  die_eof() \
  fatal(100, "unexpected eof encountered in record ",\
             nfmt_uint32(nfmt, input_line))

#define  die_format()\
  fatal(100, "error in input format in record ",\
             nfmt_uint32(nfmt, input_line))

#define  die_overflow()\
  fatal(100, "numeric overflow in record ",\
             nfmt_uint32(nfmt, input_line))


/* functions in local scope: */
static void    do_getc(char *ch);
static void    do_stdin_hdbin(struct cdbmk *C);
static void    do_stdin_cdbin(struct cdbmk *C);
static void    do_stdin_getln(struct cdbmk *C);


static
void do_getc(char *ch)
{
   switch (ioq_GETC(ioq0, ch)) {
   case (1):                   /*ok */
      break;
   case (0):                   /*eof */
      eof = 1;
      if (!eof_ok)
         die_eof();
      break;
   case (-1):                  /* io error, fallthrough: */
   default:
      fatal_syserr("failure reading input");
      break;
   }

   return;
}


/* do_stdin_hdbin()
**   process stdin as "xchange" hdb input format
**   fast:
**     uses cdbmk_ioqadd() "wirespeed" interface
**     no double copying from input
*/
static
void do_stdin_hdbin(struct cdbmk *C)
{
   char           ch;
   uint32_t       u, klen, dlen;

   /* process stdin until eof: */
   for (;;) {
      /* eof? */
      eof_ok = 1;
      do_getc(&ch);
      /* break on eof or empty line: */
      if ((eof) || (ch == '\n'))
         break;

      eof_ok = 0;
      ++input_line;

      /* klen: */
      klen = 0;
      for (;;) {
         if (ch == ':')
            break;
         if ((ch < '0') || (ch > '9'))
            die_format();
         u = klen * 10 + (ch - '0');
         if (u < klen)
            die_overflow();
         klen = u;
         do_getc(&ch);
      }

      /* dlen: */
      dlen = 0;
      for (;;) {
         do_getc(&ch);
         if (ch == '\t')
            break;
         if ((ch < '0') || (ch > '9'))
            die_format();
         u = dlen * 10 + (ch - '0');
         if (u < dlen)
            die_overflow();
         dlen = u;
      }

      /* cdbmk "wirespeed" interface: */
      if (cdbmk_addioq(C, ioq0, klen, dlen, (uchar_t *) ":", 1) == -1) {
         fatal_syserr("failure cdbmk_ioqadd()");
      }

      do_getc(&ch);
      if (ch != '\n')
         die_format();
   }

   /* here at eof: */
   return;
}


/* do_stdin_cdbin()
**   process stdin as cdb input format
**   fast:
**     uses cdbmk_ioqadd() "wirespeed" interface
**     no double copying from input
*/
static
void do_stdin_cdbin(struct cdbmk *C)
{
   char           ch;
   uint32_t       u, klen, dlen;

   /* process stdin until eof: */
   for (;;) {
      /* eof? */
      eof_ok = 1;
      do_getc(&ch);
      /* break on eof or empty line: */
      if ((eof) || (ch == '\n'))
         break;

      eof_ok = 0;
      ++input_line;

      if (ch != '+')
         die_format();

      /* klen: */
      klen = 0;
      for (;;) {
         do_getc(&ch);
         if (ch == ',')
            break;
         if ((ch < '0') || (ch > '9'))
            die_format();
         u = klen * 10 + (ch - '0');
         if (u < klen)
            die_overflow();
         klen = u;
      }

      /* dlen: */
      dlen = 0;
      for (;;) {
         do_getc(&ch);
         if (ch == ':')
            break;
         if ((ch < '0') || (ch > '9'))
            die_format();
         u = dlen * 10 + (ch - '0');
         if (u < dlen)
            die_overflow();
         dlen = u;
      }

      /* cdbmk "wirespeed" interface: */
      if (cdbmk_addioq(C, ioq0, klen, dlen, (uchar_t *) "->", 2) == -1) {
         fatal_syserr("failure cdbmk_ioqadd()");
      }

      do_getc(&ch);
      if (ch != '\n')
         die_format();
   }

   /* here at eof: */
   return;
}


/* do_stdin_getln()
**   process stdin with getln, parse key/data
**   slower: double copy of input
*/
static
void do_stdin_getln(struct cdbmk *C)
{
   dynstr_t       L = dynstr_INIT();
   char          *line;
   size_t         llen;
   uint32_t       klen, dlen;
   char           delim = arg_d[0];
   int            r;

   for (;;) {
      dynstr_CLEAR(&L);
      r = ioq_getln(ioq0, &L);
      if (r == -1) {
         fatal_syserr("failure reading input");
      } else if (r == 0) {
         /* eof: */
         break;
      }
      ++input_line;
      dynstr_chop(&L);
      line = dynstr_STR(&L);
      llen = dynstr_LEN(&L);

      if (llen == 0)
         die_format();
      /* skip commented lines: */
      if (line[0] == '#') {
         continue;
      }

      klen = cstr_pos(line, (int) delim);
      if ((klen == 0) || (line[klen] != delim)) {
         die_format();
      }

      dlen = llen - klen - 1;

      if (cdbmk_add(C,
                    (const uchar_t *) line, klen,
                    (const uchar_t *) (line + klen + 1), dlen) == -1) {
         fatal_syserr("failure cdbmk_add()");
      }
   }

   /* eof: */
   dynstr_freestr(&L);
   return;
}

/* module entry: */
int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   uint32_t       u;
   const char    *z;
   struct cdbmk   C;
   int            fd = -1, perms = -1;
   char          *outfile = NULL, *tmpfile = NULL;
   char          *fn_tmp = NULL;        /* non-NULL if allocated */

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog =
       "[-hV] [-d del | -g | -x] [-p perms] [-t tmpfile] fname <input";
   nextopt_init(&nopt, argc, argv, ":hVd:gxp:t:");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode =
       "[-h] [-d del | -g | -x] [-p perms] [-t tmpfile] fname <input";
   nextopt_init(&nopt, argc, argv, ":hd:gxp:t:");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'd':
         ++opt_g;
         arg_d = nopt.opt_arg;
         break;
      case 'g':
         ++opt_g;
         break;
      case 'x':
         ++opt_x;
         break;
      case 'p':
         z = nuscan_uint32o(&u, nopt.opt_arg);
         if (*z != '\0') {
            fatal_usage("non-numeric/octal argument found for option -", optc,
                        ": ", nopt.opt_arg);
         }
         perms = (int) u;
         break;
      case 't':
         tmpfile = nopt.opt_arg;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if (opt_g && opt_x) {
      fatal_usage("options -g and -x are mutually exclusive");
   }

   if ((argc > 0) && (argv[0] != NULL) && (*argv[0] != '\0')) {
      outfile = argv[0];
      --argc;
      ++argv;
   } else {
      fatal_usage("missing output file argument");
   }

   if (tmpfile == NULL) {
      fn_tmp = (char *) malloc(cstr_len(outfile) + cstr_len(tmpext) + 1);
      if (fn_tmp == NULL) {
         fatal_syserr("failure malloc() for naming tmpfile");
      }
      cstr_vcopy(fn_tmp, outfile, tmpext);
      tmpfile = fn_tmp;
   }

   if (perms >= 0)
      umask(0);
   else
      perms = 0666;

   unlink(tmpfile);
   fd = open(tmpfile, O_WRONLY | O_CREAT | O_EXCL, (mode_t) perms);
   if (fd == -1) {
      fatal_syserr("failure on open() for ", tmpfile);
   }

   if (cdbmk_init(&C, fd) == -1) {
      fatal_syserr("failure cdbmk_init()");
   }

   /* note: comment [*1] */
   /* branch according to input format, process stdin to eof: */
   if (opt_g) {
      /* "getln" input format: */
      do_stdin_getln(&C);
   } else if (opt_x) {
      /* "xchange" hdb input format: */
      do_stdin_hdbin(&C);
   } else {
      /* cdb input format (default): */
      do_stdin_cdbin(&C);
   }

   /* wrap it up: */
   if (cdbmk_finish(&C) == -1) {
      fatal_syserr("failure cdbmk_finish()");
   }

   if (fsync(fd) == -1) {
      fatal_syserr("failure fsync()");
   }

   if (close(fd) == -1) {
      fatal_syserr("failure close()");
   }

   if (rename(tmpfile, outfile) == -1) {
      fatal_syserr("failure rename() ", tmpfile, " to ", outfile);
   }

   /* gratuitous cleanup: */
   cdbmk_clear(&C);
   if (fn_tmp != NULL)
      free(fn_tmp);

   return 0;
}


/* eof (ezcdb_make.c) */
