/* ezcdb_stats.c
** ezcdb stats module
** wcm, 2010.05.30 - 2010.12.07
** ===
*/

/* libc: */
#include <stdlib.h>

/* unix: */
#include <unistd.h>
#include <errno.h>

/* lasagna: */
#include "buf.h"
#include "cstr.h"
#include "ioq.h"
#include "ioq_std.h"
#include "nextopt.h"
#include "nfmt.h"
#include "sysstr.h"
#include "uchar.h"
#include "upak.h"

/* lasagna cdb reader: */
#include "cdb.h"

/* ezcdb: */
#include "ezcdb.h"


#ifndef EZCDB_MODULE
/* configuration for standalone program: */
/* these globals setup in main(): */
const char    *progname = NULL;
const char    *usage_prog = NULL;
/* these globals remain NULL in standalone: */
const char    *progmode = NULL;
const char    *usage_mode = NULL;
/* entry point for standalone: */
#define  EZCDB_MAIN  main
#else
/* entry point for module: */
#define  EZCDB_MAIN  ezcdb_stats
#endif                          /* #ifndef EZCDB_MODULE */


/* additional i/o macros: */
#define vputs(...) \
  ioq_vputs(ioq1, __VA_ARGS__)


/* objects in local scope: */
/* -v: "verbose" */
static int     opt_v = 0;

/* functions in local scope: */
static int     do_stats(struct cdb *C);


/* do_stats()
**   scan cdb and output statistics
**   abort on failure
*/
static
int do_stats(struct cdb *C)
{
   uchar_t        nbuf[8];
   uint32_t       nrecs = 0;
   uchar_t       *pbuf = NULL;  /* allocated */
   uchar_t       *tbuf = NULL;  /* allocated */
   uchar_t        tbuf_size = 0;
   uint32_t       tbase, tslots;
   uint32_t       klen, klen_min = 0, klen_max = 0, klen_sum = 0, klen_avg;
   uint32_t       dlen, dlen_min = 0, dlen_max = 0, dlen_sum = 0, dlen_avg;
   uint32_t       rpos = 0;
   uint32_t       dist, distv[11], dist_max = 0, dist_cum = 0;
   uint32_t       dist_max_rpos = cdb_rbase(C);
   uint32_t       dw_cum = 0, dw_int, dw_frac;
   size_t         t, slot;
   uint32_t       h, want;
   uint32_t       sum_slots = 0, min_slots = 0, max_slots = 0;
   uchar_t       *kbuf = NULL;
   char           nfmt[4][NFMT_SIZE];
   int            i, r, w;

   pbuf = (uchar_t *) malloc(CDB_NTABS * 8);
   if (pbuf == NULL) {
      fatal_syserr("failure allocating buffer for hash table pointers");
   }

   if (cdb_read(C, pbuf, (CDB_NTABS * 8), cdb_pbase(C)) == -1) {
      fatal_syserr("failure reading hash table pointers");
   }

   /* sequential scan all records in record block: */
   /* XXX
    **   cdb_seq*() interface is not buffered
    **   so this is not an efficient way of doing business
    */
   for (r = cdb_seqinit(C); r != 0; r = cdb_seqnext(C)) {
      if (r == -1) {
         fatal_syserr("failure during cdb_seq()");
      }

      klen = cdb_klen(C);
      klen_sum += klen;
      if ((klen_min == 0) || (klen_min > klen))
         klen_min = klen;
      if (klen_max < klen)
         klen_max = klen;

      dlen = cdb_dlen(C);
      dlen_sum += dlen;
      if ((dlen_min == 0) || (dlen_min > dlen))
         dlen_min = dlen;
      if (dlen_max < dlen)
         dlen_max = dlen;

      ++nrecs;
   }

   /* initialize stats: */
   klen_avg = nrecs ? (klen_sum / nrecs) : 0;
   dlen_avg = nrecs ? (dlen_sum / nrecs) : 0;

   /* scan all hash subtables: */
   /* note: "distance" algorithm derived from tinycdb by michael tokarev */
   buf_zero(distv, sizeof distv);
   for (t = 0; t < CDB_NTABS; ++t) {
      tbase = upak32_unpack(pbuf + (t * 8));
      tslots = upak32_unpack(pbuf + (t * 8) + 4);

      if (tslots == 0)
         continue;

      /* suck in this hash table (profligate memory usage!): */
      if (tbuf_size < (tslots * 8)) {
         if ((tbuf = (uchar_t *) realloc(tbuf, tslots * 8)) == NULL) {
            fatal_syserr("failure realloc() hash table index");
         }
         tbuf_size = tslots * 8;
      }

      if (cdb_read(C, tbuf, tslots * 8, tbase) == -1) {
         fatal_syserr("failure cdb_read()");
      }

      sum_slots += tslots;
      if ((min_slots == 0) || (min_slots > tslots))
         min_slots = tslots;
      if (max_slots < tslots)
         max_slots = tslots;

      /* scan slots in hash table t: */
      for (slot = 0; slot < tslots; ++slot) {
         rpos = upak32_unpack(tbuf + (slot * 8) + 4);
         /* continue if this slot empty: */
         if (rpos == 0)
            continue;
         /* hash value found here: */
         h = upak32_unpack(tbuf + (slot * 8));
         /* slot in table this hash value wants: */
         want = cdb_SLOT(h, tslots);
         /* distance: */
         if (want == slot) {
            dist = 0;
         } else {
            dist = (want < slot) ? (slot - want) : ((tslots - want) + slot);
         }
         /* increment counters: */
         dw_cum += dist;
         if (dist < 10)
            ++distv[dist];
         else
            ++distv[10];
         /* maximum distance: */
         if (dist_max < dist) {
            dist_max = dist;
            dist_max_rpos = rpos;
         }
      }
   }
   /* release allocated buffers: */
   free(tbuf);
   free(pbuf);

   /* report: */
   vputs("cdb total records:  ", nfmt_uint32(nfmt[0], nrecs), "\n");
   if (opt_v) {
      vputs("cdb reader mmap() enabled:  ", cdb_map(C) ? "yes" : "no", "\n");
   }
   vputs("key length (min/avg/max):  ",
         nfmt_uint32(nfmt[0], klen_min), "/",
         nfmt_uint32(nfmt[1], klen_avg), "/",
         nfmt_uint32(nfmt[2], klen_max), "\n");
   vputs("data length (min/avg/max):  ",
         nfmt_uint32(nfmt[0], dlen_min), "/",
         nfmt_uint32(nfmt[1], dlen_avg), "/",
         nfmt_uint32(nfmt[2], dlen_max), "\n");
   vputs("hash subtables:  ", nfmt_uint32(nfmt[0], CDB_NTABS), "\n");
   vputs("slots/subtable (min/avg/max):  ",
         nfmt_uint32(nfmt[0], min_slots), "/",
         nfmt_uint32(nfmt[1], (sum_slots / CDB_NTABS)), "/",
         nfmt_uint32(nfmt[2], max_slots), "\n");
   vputs("hash insert collisions:  ",
         nfmt_uint32(nfmt[0], nrecs - distv[0]), " (",
         nfmt_uint32(nfmt[1],
                     nrecs ? (((nrecs - distv[0]) * 100) / nrecs) : 0),
         "%)\n");
   vputs("hash search distances:\n");
   /* field width for the following table: */
   w = nfmt_uint32_(NULL, nrecs) + 1;
   for (i = 0; i < 10; ++i) {
      dist_cum += distv[i];
      vputs("  d", nfmt_uint32(nfmt[0], i), ":  ",
            nfmt_uint32_pad(nfmt[1], distv[i], w), "  ",
            nfmt_uint32_pad(nfmt[2], nrecs ? ((distv[i] * 100) / nrecs) : 0,
                            3), "%  (", nfmt_uint32_pad(nfmt[3],
                                                        nrecs
                                                        ? ((dist_cum * 100) /
                                                           nrecs) : 0, 3),
            "%)\n");
   }
   dist_cum += distv[10];
   vputs("  >9:  ",
         nfmt_uint32_pad(nfmt[1], distv[10], w), "  ",
         nfmt_uint32_pad(nfmt[2], nrecs ? ((distv[10] * 100) / nrecs) : 0, 3),
         "%  (", nfmt_uint32_pad(nfmt[3],
                                 nrecs ? ((dist_cum * 100) / nrecs) : 0, 3),
         "%)\n");

   vputs("maximum search distance:  ", nfmt_uint32(nfmt[0], dist_max), "\n");

   dw_int = nrecs ? (dw_cum / nrecs) : 0;
   dw_frac = nrecs ? (((dw_cum * 1000) / nrecs) - (dw_int * 1000)) : 0;
   vputs("average search distance:  ",
         nfmt_uint32(nfmt[0], dw_int), ".", nfmt_uint32(nfmt[1], dw_frac),
         "\n");

   if (opt_v) {
      if (nrecs > 0) {
         /* get key found at maximum search distance: */
         if (cdb_read(C, nbuf, 8, dist_max_rpos) == -1) {
            fatal_syserr("failure cdb_read()");
         }
         klen = upak32_unpack(nbuf);
         kbuf = (uchar_t *) malloc(klen + 1);
         if (kbuf == NULL) {
            fatal_syserr("failure malloc() for key buffer");
         }
         if (cdb_read(C, kbuf, klen, dist_max_rpos + 8) == -1) {
            fatal_syserr("failure cdb_read()");
         }
         kbuf[klen] = '\0';
         vputs("key at maximum search distance:  ", kbuf, "\n");
         free(kbuf);
      } else {
         vputs("key at maximum search distance: (no records)\n");
      }
   }

   /* all done: */
   ioq_flush(ioq1);
   return 0;
}


int EZCDB_MAIN(int argc, char *argv[])
{
   nextopt_t      nopt;
   char           opt;
   struct cdb     C;
   int            is_stdin = 0;
   int            e;

#ifndef EZCDB_MODULE
/* standalone options/initialization: */
   usage_prog = "[-hV] [-v] [cdb]";
   nextopt_init(&nopt, argc, argv, ":hVv");
   progname = nextopt_progname(&nopt);
#else
/* module options/initialization: */
   usage_mode = "[-h] [-v] [cdb]";
   nextopt_init(&nopt, argc, argv, ":hv");
#endif

   while ((opt = nextopt(&nopt))) {
      char           optc[2] = { nopt.opt_got, '\0' };
      switch (opt) {
      case 'h':
         usage();
         die(0);
         break;
      case 'V':
         version();
         die(0);
         break;
      case 'v':
         ++opt_v;
         break;
      case ':':
         fatal_usage("missing argument for option -", optc);
         break;
      case '?':
         if (nopt.opt_got != '?') {
            fatal_usage("invalid option -", optc);
         }
         /* else fallthrough: */
      default:
         die_usage();
         break;
      }
   }

   argc -= nopt.arg_ndx;
   argv += nopt.arg_ndx;

   if ((argc > 0) && (argv[0] != NULL)) {
      /* cdb file argument: */
      if (cdb_open(&C, argv[0]) == -1) {
         fatal_syserr("failure initializing cdb from ", argv[0]);
      }
   } else {
      /* read from stdin: */
      if (cdb_init(&C, 0) == -1) {
         fatal_syserr("failure initializing cdb on stdin");
      }
      ++is_stdin;
   }

   /* note: do_stats() just aborts on failure */
   e = do_stats(&C);

   /* cleanup/close: */
   if (is_stdin) {
      cdb_clear(&C);
   } else {
      cdb_close(&C);
   }

   return e;
}


/* eof (ezcdb_stats.c) */
